#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ~~ Imports ~~ ────────────────────────────────────────────────────────────────
from typing import List

# ~~ Exports ~~ ────────────────────────────────────────────────────────────────
__all__: List[str] = [
    "random_hcu",
    "random_hoe",
    "random_woe",
    "random_hue",
    "random_wue",
    "random_mce",
    "random_mre",
    "random_lce",
    "random_lre",
]
# ──────────────────────────────────────────────────────────────────────────────
from .api import random_cue as random_hcu
from .api import random_goe as random_hoe
from .api import random_goe as random_woe
from .api import random_gue as random_hue
from .api import random_gue as random_wue
from .api import random_jce as random_mce
from .api import random_jre as random_mre
from .api import random_wce as random_lce
from .api import random_wre as random_lre



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# ~~ Imports ~~ ────────────────────────────────────────────────────────────────
import itertools
import math
from typing import List
from typing import Optional
from typing import Tuple

import torch as th
from torch import Tensor

from .types import c2r_map

# ~~ Exports ~~ ────────────────────────────────────────────────────────────────
__all__: List[str] = [
    "random_obs_cgi",  # Observable as linear combination of Generalized Gell-Mann matrices
]


# ~~ Functions ~~ ──────────────────────────────────────────────────────────────


def _gen_gmm(
    size: int,
    dtype=th.cdouble,
    device: Optional[th.device] = None,
) -> List[Tensor]:
    matrices: List[Tensor] = []

    for i, j in itertools.combinations(range(size), 2):
        symmetric: Tensor = th.zeros((size, size), dtype=dtype, device=device)
        symmetric[i, j] = 1.0 / math.sqrt(2.0)
        symmetric[j, i] = 1.0 / math.sqrt(2.0)
        matrices.append(symmetric)

        antisymmetric: Tensor = th.zeros((size, size), dtype=dtype, device=device)
        antisymmetric[i, j] = -1j / math.sqrt(2.0)  # type: ignore
        antisymmetric[j, i] = 1j / math.sqrt(2.0)  # type: ignore
        matrices.append(antisymmetric)

    for k in range(1, size):
        diag_matrix: Tensor = th.zeros((size, size), dtype=dtype, device=device)
        coeff: float = 1.0 / math.sqrt(k * (k + 1.0))
        for i in range(k):
            diag_matrix[i, i] = coeff
        diag_matrix[k, k] = -k * coeff
        matrices.append(diag_matrix)

    return matrices


def random_obs_cgi(
    size: int,
    coeff_low: float,
    coeff_upp: float,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Tuple[int, ...] = (),
) -> Tensor:
    gell_mann_matrices: List[Tensor] = _gen_gmm(size, dtype=dtype, device=device)

    batch_size: Tuple[int, ...] = batch_shape if batch_shape else (1,)
    random_coeffs: Tensor = (coeff_upp - coeff_low) * th.rand(
        *batch_size, size * size - 1, dtype=c2r_map[dtype], device=device
    ) + coeff_low

    result: Tensor = (
        sum(  # type: ignore
            random_coeffs[..., i, None, None] * m
            for i, m in enumerate(gell_mann_matrices)
        )
        + th.eye(size, dtype=dtype, device=device) / size
    )
    return result.squeeze(0) if batch_shape == () else result  # type: ignore



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# ~~ Imports ~~ ────────────────────────────────────────────────────────────────
from typing import List

import torch as th

# ~~ Exports ~~ ────────────────────────────────────────────────────────────────
__all__: List[str] = [
    "check_dtype",
    "check_size",
    "check_sigma",
    "check_bounds",
]


# ~~ Functions ~~ ──────────────────────────────────────────────────────────────
def check_dtype(
    dtype: th.dtype,
    valid_dtypes: List[th.dtype],
) -> None:
    if dtype not in valid_dtypes:
        raise ValueError(f"Invalid `dtype`: {dtype}. Must be one of {valid_dtypes}")


def check_size(size: int, disallow_equality: bool = False) -> None:
    if size < (1 + disallow_equality):
        raise ValueError(f"Invalid `size`: {size}. Must be > {disallow_equality}.")


def check_sigma(sigma: float) -> None:
    if sigma < 0:
        raise ValueError(f"Invalid `sigma`: {sigma}. Must be >= 0.")


def check_bounds(lower: float, upper: float) -> None:
    if lower > upper:
        raise ValueError("Invalid limits: `lower` must be <= `upper`")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# ~~ Imports ~~ ────────────────────────────────────────────────────────────────
from math import sqrt as msqrt
from typing import List
from typing import Optional
from typing import Tuple

import torch as th
from torch import Tensor

# ~~ Exports ~~ ────────────────────────────────────────────────────────────────

__all__: List[str] = [
    "random_coe",  # Circular Orthogonal Ensemble
    "random_cue",  # Circular Unitary (Haar Uniform) Ensemble
    "random_gce",  # Ginibre Complex Ensemble
    "random_goe",  # Gaussian (Hermite, or Wigner) Orthogonal Ensemble
    "random_gre",  # Ginibre Real Ensemble
    "random_gue",  # Gaussian (Hermite, or Wigner) Unitary Ensemble
    "random_jce",  # Jacobi (MANOVA) Complex Ensemble
    "random_jre",  # Jacobi (MANOVA) Real Ensemble
    "random_wce",  # Wishart (Laguerre) Complex Ensemble
    "random_wre",  # Wishart (Laguerre) Real Ensemble
]

# ~~ Functions ~~ ──────────────────────────────────────────────────────────────


def random_gre(
    size: int,
    nnorm: bool = False,
    dtype: th.dtype = th.double,
    device: Optional[th.device] = None,
    batch_shape: Tuple[int, ...] = (),
):
    x: Tensor = th.randn(*batch_shape, size, size, dtype=dtype, device=device)
    if nnorm:
        x: Tensor = x / msqrt(size)
    return x


def random_gce(
    size: int,
    nnorm: bool = False,
    cnorm: bool = True,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Tuple[int, ...] = (),
):
    x: Tensor = th.randn(*batch_shape, size, size, dtype=dtype, device=device)
    if nnorm and (not cnorm):
        x: Tensor = x / msqrt(size / 2)
    else:
        if nnorm:
            x: Tensor = x / msqrt(size)
        if not cnorm:
            x: Tensor = x * msqrt(2)
    return x


def random_cue(
    size: int,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Tuple[int, ...] = (),
) -> Tensor:
    a: Tensor = random_gce(
        size=size, dtype=dtype, device=device, batch_shape=batch_shape
    )
    q, r = th.linalg.qr(a)
    d: Tensor = r.diagonal(dim1=-2, dim2=-1)
    return q @ th.diag_embed(d / th.abs(d))


def random_coe(
    size: int,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Tuple[int, ...] = (),
) -> Tensor:
    x: Tensor = random_cue(
        size=size, dtype=dtype, device=device, batch_shape=batch_shape
    )
    return x.transpose(-2, -1) @ x


def random_gue(
    size: int,
    sigma: float,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Tuple[int, ...] = (),
) -> Tensor:
    x: Tensor = (
        random_gce(size=size, dtype=dtype, device=device, batch_shape=batch_shape)
        * sigma
    )
    return (x + x.transpose(-2, -1).conj()) / msqrt(2)


def random_goe(
    size: int,
    sigma: float,
    dtype: th.dtype = th.double,
    device: Optional[th.device] = None,
    batch_shape: Tuple[int, ...] = (),
) -> Tensor:
    x: Tensor = (
        random_gre(size=size, dtype=dtype, device=device, batch_shape=batch_shape)
        * sigma
    )
    return (x + x.transpose(-2, -1)) / msqrt(2)


def random_wre(
    size_n: int,
    sigma: float,
    size_m: Optional[int] = None,
    dtype: th.dtype = th.double,
    device: Optional[th.device] = None,
    batch_shape: Tuple[int, ...] = (),
) -> Tensor:
    actual_size_m: int = size_m or size_n
    x: Tensor = (
        th.randn(*batch_shape, size_n, actual_size_m, dtype=dtype, device=device)
        * sigma
    )
    return x @ x.transpose(-2, -1)


def random_wce(
    size_n: int,
    sigma: float,
    size_m: Optional[int] = None,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Tuple[int, ...] = (),
) -> Tensor:
    actual_size_m: int = size_m or size_n
    x: Tensor = (
        th.randn(*batch_shape, size_n, actual_size_m, dtype=dtype, device=device)
        * sigma
    )
    return x @ x.transpose(-2, -1).conj()


def random_jre(
    size_n: int,
    size_m1: Optional[int] = None,
    size_m2: Optional[int] = None,
    dtype: th.dtype = th.double,
    device: Optional[th.device] = None,
    batch_shape: Tuple[int, ...] = (),
) -> Tensor:
    actual_size_m1: int = size_m1 or size_n
    actual_size_m2: int = size_m2 or size_n
    x: Tensor = th.randn(
        *batch_shape, size_n, actual_size_m1, dtype=dtype, device=device
    )
    a1: Tensor = x @ x.transpose(-2, -1)
    y: Tensor = th.randn(
        *batch_shape, size_n, actual_size_m2, dtype=dtype, device=device
    )
    a2: Tensor = a1 + y @ y.transpose(-2, -1)
    return a1 @ th.linalg.inv(a2)


def random_jce(
    size_n: int,
    size_m1: Optional[int] = None,
    size_m2: Optional[int] = None,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Tuple[int, ...] = (),
) -> Tensor:
    actual_size_m1: int = size_m1 or size_n
    actual_size_m2: int = size_m2 or size_n
    x: Tensor = th.randn(
        *batch_shape, size_n, actual_size_m1, dtype=dtype, device=device
    )
    a1: Tensor = x @ x.transpose(-2, -1).conj()
    y: Tensor = th.randn(
        *batch_shape, size_n, actual_size_m2, dtype=dtype, device=device
    )
    a2: Tensor = a1 + y @ y.transpose(-2, -1).conj()
    return a1 @ th.linalg.inv(a2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# ~~ Imports ~~ ────────────────────────────────────────────────────────────────
from typing import List
from typing import Optional
from typing import Tuple

import torch as th
from torch import Tensor

from .auxiliary import check_dtype
from .auxiliary import check_sigma
from .auxiliary import check_size
from .impl import random_coe as _random_coe
from .impl import random_cue as _random_cue
from .impl import random_gce as _random_gce
from .impl import random_goe as _random_goe
from .impl import random_gre as _random_gre
from .impl import random_gue as _random_gue
from .impl import random_jce as _random_jce
from .impl import random_jre as _random_jre
from .impl import random_wce as _random_wce
from .impl import random_wre as _random_wre
from .types import complex_dtypes
from .types import real_dtypes


# ~~ Exports ~~ ────────────────────────────────────────────────────────────────

__all__: List[str] = [
    "random_coe",  # Circular Orthogonal Ensemble
    "random_cue",  # Circular Unitary (Haar Uniform) Ensemble
    "random_gce",  # Ginibre Complex Ensemble
    "random_goe",  # Gaussian (Hermite, or Wigner) Orthogonal Ensemble
    "random_gre",  # Ginibre Real Ensemble
    "random_gue",  # Gaussian (Hermite, or Wigner) Unitary Ensemble
    "random_jce",  # Jacobi (MANOVA) Complex Ensemble
    "random_jre",  # Jacobi (MANOVA) Real Ensemble
    "random_wce",  # Wishart (Laguerre) Complex Ensemble
    "random_wre",  # Wishart (Laguerre) Real Ensemble
]


# ~~ Functions ~~ ──────────────────────────────────────────────────────────────


def random_gre(
    size: int,
    nnorm: bool = False,
    dtype: th.dtype = th.double,
    device: Optional[th.device] = None,
    batch_shape: Optional[Tuple[int, ...]] = None,
) -> Tensor:
    """
    Generate a random matrix (or a batch thereof) from the Ginibre Real Ensemble (GRE).

    Parameters
    ----------
    size : int
        The size of the square matrix.
    nnorm : bool, optional
        Normalize the matrix by the square root of its size. Default is False.
    dtype : torch.dtype
        The data type. Default is torch.double.
    device : torch.device, optional
        The device for the tensor. Default is None.
    batch_shape : tuple of ints, optional
        The batch shape for generating multiple matrices. Default is None.

    Returns
    -------
    Tensor
        A random GRE matrix of shape (*batch_shape, size, size).
    """
    check_size(size)
    check_dtype(dtype, real_dtypes)
    bs = () if batch_shape is None else batch_shape
    return _random_gre(
        size=size, nnorm=nnorm, dtype=dtype, device=device, batch_shape=bs
    )


def random_gce(
    size: int,
    nnorm: bool = False,
    cnorm: bool = True,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Optional[Tuple[int, ...]] = None,
) -> Tensor:
    """
    Generate a random matrix (or a batch thereof) from the Ginibre Complex Ensemble (GCE).

    Parameters
    ----------
    size : int
        The size of the square matrix.
    nnorm : bool, optional
        Normalize the matrix by the square root of its size. Default is False.
    cnorm : bool, optional
        Normalize the matrix by the square root of 2. Default is True.
    dtype : torch.dtype
        The data type. Default is torch.cdouble.
    device : torch.device, optional
        The device for the tensor. Default is None.
    batch_shape : tuple of ints, optional
        The batch shape for generating multiple matrices. Default is None.

    Returns
    -------
    Tensor
        A random GCE matrix of shape (*batch_shape, size, size).
    """
    check_size(size)
    check_dtype(dtype, complex_dtypes)
    bs = () if batch_shape is None else batch_shape
    return _random_gce(
        size=size, nnorm=nnorm, cnorm=cnorm, dtype=dtype, device=device, batch_shape=bs
    )


# noinspection DuplicatedCode
def random_cue(
    size: int,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Optional[Tuple[int, ...]] = None,
) -> Tensor:
    """
    Generate a random unitary matrix (or a batch thereof) from the Circular Unitary Ensemble (CUE),
    also known as the Uniform Haar measure on the unitary group.

    Parameters
    ----------
    size : int
        The size of the square matrix.
    dtype : torch.dtype
        The data type. Default is torch.cdouble.
    device : torch.device, optional
        The device for the tensor. Default is None.
    batch_shape : tuple of ints, optional
        The batch shape for generating multiple matrices. For example, batch_shape=(B,)
        returns a tensor of shape (B, size, size). Default is None (i.e. a single matrix).

    Returns
    -------
    Tensor
        A random unitary matrix of shape (*batch_shape, size, size).

    Raises
    ------
    ValueError
        If size is less than 1 or dtype is invalid.
    """
    check_size(size)
    check_dtype(dtype, complex_dtypes)
    bs = () if batch_shape is None else batch_shape
    return _random_cue(size=size, dtype=dtype, device=device, batch_shape=bs)


def random_coe(
    size: int,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Optional[Tuple[int, ...]] = None,
) -> Tensor:
    """
    Generate a random matrix (or a batch thereof) from the Circular Orthogonal Ensemble (COE).

    Parameters
    ----------
    size : int
        The size of the square matrix.
    dtype : torch.dtype
        The data type. Default is torch.cdouble.
    device : torch.device, optional
        The device for the tensor. Default is None.
    batch_shape : tuple of ints, optional
        The batch shape for generating multiple matrices. Default is None.

    Returns
    -------
    Tensor
        A random COE matrix of shape (*batch_shape, size, size).
    """
    check_size(size)
    check_dtype(dtype, complex_dtypes)
    bs = () if batch_shape is None else batch_shape
    return _random_coe(size=size, dtype=dtype, device=device, batch_shape=bs)


def random_gue(
    size: int,
    sigma: float,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Optional[Tuple[int, ...]] = None,
) -> Tensor:
    """
    Generate a matrix (or a batch thereof) from the Gaussian Unitary Ensemble (GUE) of scale sigma,
    also known as the Hermite (or Wigner) Unitary Ensemble.

    Parameters
    ----------
    size : int
        The size of the square matrix.
    sigma : float
        The scale parameter.
    dtype : torch.dtype
        The data type. Default is torch.cdouble.
    device : torch.device, optional
        The device for the tensor. Default is None.
    batch_shape : tuple of ints, optional
        The batch shape for generating multiple matrices. Default is None.

    Returns
    -------
    Tensor
        A random GUE matrix of shape (*batch_shape, size, size).
    """
    check_size(size)
    check_dtype(dtype, complex_dtypes)
    check_sigma(sigma)
    bs = () if batch_shape is None else batch_shape
    return _random_gue(
        size=size, sigma=sigma, dtype=dtype, device=device, batch_shape=bs
    )


def random_goe(
    size: int,
    sigma: float,
    dtype: th.dtype = th.double,
    device: Optional[th.device] = None,
    batch_shape: Optional[Tuple[int, ...]] = None,
) -> Tensor:
    """
    Generate a matrix (or a batch thereof) from the Gaussian Orthogonal Ensemble (GOE) of scale sigma,
    also known as the Hermite (or Wigner) Orthogonal Ensemble.

    Parameters
    ----------
    size : int
        The size of the square matrix.
    sigma : float
        The scale parameter.
    dtype : torch.dtype
        The data type. Default is torch.double.
    device : torch.device, optional
        The device for the tensor. Default is None.
    batch_shape : tuple of ints, optional
        The batch shape for generating multiple matrices. Default is None.

    Returns
    -------
    Tensor
        A random GOE matrix of shape (*batch_shape, size, size).
    """
    check_size(size)
    check_dtype(dtype, real_dtypes)
    check_sigma(sigma)
    bs = () if batch_shape is None else batch_shape
    return _random_goe(
        size=size, sigma=sigma, dtype=dtype, device=device, batch_shape=bs
    )


# noinspection DuplicatedCode
def random_wre(
    size_n: int,
    sigma: float,
    size_m: Optional[int] = None,
    dtype: th.dtype = th.double,
    device: Optional[th.device] = None,
    batch_shape: Optional[Tuple[int, ...]] = None,
) -> Tensor:
    """
    Generate a matrix (or a batch thereof) from the Wishart Real Ensemble (WRE), also known as the Laguerre Ensemble.

    Parameters
    ----------
    size_n : int
        The number of rows.
    sigma : float
        The scale parameter.
    size_m : int, optional
        The number of columns. If None, defaults to size_n.
    dtype : torch.dtype
        The data type. Default is torch.double.
    device : torch.device, optional
        The device for the tensor. Default is None.
    batch_shape : tuple of ints, optional
        The batch shape for generating multiple matrices. Default is None.

    Returns
    -------
    Tensor
        A random WRE matrix of shape (*batch_shape, size_n, size_m).
    """
    check_size(size_n)
    check_sigma(sigma)
    if size_m is not None:
        check_size(size_m)
    bs = () if batch_shape is None else batch_shape
    return _random_wre(
        size_n=size_n,
        sigma=sigma,
        size_m=size_m,
        dtype=dtype,
        device=device,
        batch_shape=bs,
    )


# noinspection DuplicatedCode
def random_wce(
    size_n: int,
    sigma: float,
    size_m: Optional[int] = None,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Optional[Tuple[int, ...]] = None,
) -> Tensor:
    """
    Generate a matrix (or a batch thereof) from the Wishart Complex Ensemble (WCE), also known as the Laguerre Ensemble.

    Parameters
    ----------
    size_n : int
        The number of rows.
    sigma : float
        The scale parameter.
    size_m : int, optional
        The number of columns. If None, defaults to size_n.
    dtype : torch.dtype
        The data type. Default is torch.cdouble.
    device : torch.device, optional
        The device for the tensor. Default is None.
    batch_shape : tuple of ints, optional
        The batch shape for generating multiple matrices. Default is None.

    Returns
    -------
    Tensor
        A random WCE matrix of shape (*batch_shape, size_n, size_m).
    """
    check_size(size_n)
    check_sigma(sigma)
    if size_m is not None:
        check_size(size_m)
    bs = () if batch_shape is None else batch_shape
    return _random_wce(
        size_n=size_n,
        sigma=sigma,
        size_m=size_m,
        dtype=dtype,
        device=device,
        batch_shape=bs,
    )


# noinspection DuplicatedCode
def random_jre(
    size_n: int,
    size_m1: Optional[int] = None,
    size_m2: Optional[int] = None,
    dtype: th.dtype = th.double,
    device: Optional[th.device] = None,
    batch_shape: Optional[Tuple[int, ...]] = None,
) -> Tensor:
    """
    Generate a matrix (or a batch thereof) from the Jacobi Real Ensemble (JRE), also known as the MANOVA Ensemble.

    Parameters
    ----------
    size_n : int
        The number of rows.
    size_m1 : int, optional
        The number of columns of the first matrix. If None, defaults to size_n.
    size_m2 : int, optional
        The number of columns of the second matrix. If None, defaults to size_n.
    dtype : torch.dtype
        The data type. Default is torch.double.
    device : torch.device, optional
        The device for the tensor. Default is None.
    batch_shape : tuple of ints, optional
        The batch shape for generating multiple matrices. Default is None.

    Returns
    -------
    Tensor
        A random JRE matrix of shape (*batch_shape, size_n, size_n).
    """
    check_size(size_n)
    if size_m1 is not None:
        check_size(size_m1)
    if size_m2 is not None:
        check_size(size_m2)
    check_dtype(dtype, real_dtypes)
    bs = () if batch_shape is None else batch_shape
    return _random_jre(
        size_n=size_n,
        size_m1=size_m1,
        size_m2=size_m2,
        dtype=dtype,
        device=device,
        batch_shape=bs,
    )


# noinspection DuplicatedCode
def random_jce(
    size_n: int,
    size_m1: Optional[int] = None,
    size_m2: Optional[int] = None,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Optional[Tuple[int, ...]] = None,
) -> Tensor:
    """
    Generate a matrix (or a batch thereof) from the Jacobi Complex Ensemble (JCE), also known as the MANOVA Ensemble.

    Parameters
    ----------
    size_n : int
        The number of rows.
    size_m1 : int, optional
        The number of columns of the first matrix. If None, defaults to size_n.
    size_m2 : int, optional
        The number of columns of the second matrix. If None, defaults to size_n.
    dtype : torch.dtype
        The data type. Default is torch.cdouble.
    device : torch.device, optional
        The device for the tensor. Default is None.
    batch_shape : tuple of ints, optional
        The batch shape for generating multiple matrices. Default is None.

    Returns
    -------
    Tensor
        A random JCE matrix of shape (*batch_shape, size_n, size_n).
    """
    check_size(size_n)
    if size_m1 is not None:
        check_size(size_m1)
    if size_m2 is not None:
        check_size(size_m2)
    check_dtype(dtype, complex_dtypes)
    bs = () if batch_shape is None else batch_shape
    return _random_jce(
        size_n=size_n,
        size_m1=size_m1,
        size_m2=size_m2,
        dtype=dtype,
        device=device,
        batch_shape=bs,
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ~~ Imports ~~ ────────────────────────────────────────────────────────────────
from collections.abc import Callable
from typing import List
from typing import Optional
from typing import Tuple

import torch as th
from torch import Tensor

from .core import batched_outer
from .impl import random_cue as _random_cue
from .impl import random_gce as _random_gce
from .impl import random_gue as _random_gue

# ~~ Exports ~~ ────────────────────────────────────────────────────────────────
__all__: List[str] = [
    "random_rho_hs",
    "random_rho_bh",
    "random_obs_gue",
    "random_obs_csu",
    "random_rho_pure",
]


# ~~ Functions ~~ ──────────────────────────────────────────────────────────────


def random_rho_hs(
    size: int,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Tuple[int, ...] = (),
):
    x: Tensor = _random_gce(
        size=size, dtype=dtype, device=device, batch_shape=batch_shape
    )
    aad: Tensor = x @ x.transpose(-2, -1).conj()
    return aad / aad.diagonal(offset=0, dim1=-2, dim2=-1).sum(-1)


def random_rho_bh(
    size: int,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Tuple[int, ...] = (),
):
    u: Tensor = _random_cue(
        size=size, dtype=dtype, device=device, batch_shape=batch_shape
    )
    a: Tensor = _random_gce(
        size=size, dtype=dtype, device=device, batch_shape=batch_shape
    )
    beye: Tensor = th.diag_embed(
        th.ones(*batch_shape, size, dtype=dtype, device=device)
    )
    x: Tensor = (beye + u) @ a
    aad: Tensor = x @ x.transpose(-2, -1).conj()
    return aad / aad.diagonal(offset=0, dim1=-2, dim2=-1).sum(-1)


def random_obs_gue(
    size: int,
    sigma: float,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Tuple[int, ...] = (),
) -> Tensor:
    return _random_gue(
        size=size, sigma=sigma, dtype=dtype, device=device, batch_shape=batch_shape
    )


def random_obs_csu(
    size: int,
    evdist: Callable[..., Tensor],
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Tuple[int, ...] = (),
) -> Tensor:
    u: Tensor = _random_cue(
        size=size, dtype=dtype, device=device, batch_shape=batch_shape
    )
    ev: Tensor = evdist(*batch_shape, size, dtype=dtype, device=device)
    return u @ th.diag_embed(ev) @ u.transpose(-2, -1).conj()


def random_rho_pure(
    size: int,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Tuple[int, ...] = (),
    *,
    bo_einsum: bool = False,
):
    p: Tensor = th.randn(*batch_shape, size, dtype=dtype, device=device)
    p = p / th.linalg.norm(p, dim=-1, keepdim=True)
    if batch_shape == ():
        return th.outer(p, p.conj())
    return batched_outer(p, p.conj(), use_einsum=bo_einsum)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ~~ Imports ~~ ────────────────────────────────────────────────────────────────
from typing import List

# ~~ Exports ~~ ────────────────────────────────────────────────────────────────
__all__: List[str] = [
    "random_coe",
    "random_cue",
    "random_gce",
    "random_goe",
    "random_gre",
    "random_gue",
    "random_jce",
    "random_jre",
    "random_obs_csu",
    "random_obs_gue",
    "random_rho_bh",
    "random_rho_hs",
    "random_wce",
    "random_wre",
    "random_obs_cgi",
    "random_rho_pure",
]

# ──────────────────────────────────────────────────────────────────────────────
# From main API
from .api import random_coe
from .api import random_cue
from .api import random_gce
from .api import random_goe
from .api import random_gre
from .api import random_gue
from .api import random_jce
from .api import random_jre
from .api import random_wce
from .api import random_wre

# From quantum-specific API
from .quantum_api import random_rho_hs
from .quantum_api import random_rho_bh
from .quantum_api import random_obs_csu
from .quantum_api import random_obs_gue
from .quantum_api import random_obs_cgi
from .quantum_api import random_rho_pure

# From Aliases
from .aliases import random_hcu
from .aliases import random_hoe
from .aliases import random_woe
from .aliases import random_hue
from .aliases import random_wue
from .aliases import random_mce
from .aliases import random_mre
from .aliases import random_lce
from .aliases import random_lre



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#  Copyright (c) 2025 Emanuele Ballarin <emanuele@ballarin.cc>
#  Released under the terms of the MIT License
#  (see: https://url.ballarin.cc/mitlicense)
#
# ------------------------------------------------------------------------------
import os

from setuptools import find_packages
from setuptools import setup


def read(fname):
    with open(os.path.join(os.path.dirname(__file__), fname)) as f:
        return f.read().strip()


PACKAGENAME: str = "thrmt"

setup(
    name=PACKAGENAME,
    version="0.0.10",
    author="Emanuele Ballarin",
    author_email="emanuele@ballarin.cc",
    url="https://github.com/emaballarin/thrmt",
    description="Torched Random Matrix Theory",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    keywords=[
        "Deep Learning",
        "Differentiable Programming",
        "Machine Learning",
        "PyTorch",
        "Random Matrices",
        "Random Matrix Theory",
    ],
    license="MIT",
    packages=[
        package for package in find_packages() if package.startswith(PACKAGENAME)
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Topic :: Scientific/Engineering :: Mathematics",
        "Environment :: Console",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires=">=3.10",
    install_requires=["torch>=2"],
    include_package_data=False,
    zip_safe=True,
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# ~~ Imports ~~ ────────────────────────────────────────────────────────────────
from collections.abc import Callable
from typing import List
from typing import Optional
from typing import Tuple

import torch as th
from torch import Tensor

from .auxiliary import check_bounds
from .auxiliary import check_dtype
from .auxiliary import check_size
from .gellmann_impl import random_obs_cgi as _random_obs_cgi
from .quantum_impl import random_obs_csu as _random_obs_csu
from .quantum_impl import random_obs_gue as _random_obs_gue
from .quantum_impl import random_rho_bh as _random_rho_bh
from .quantum_impl import random_rho_hs as _random_rho_hs
from .quantum_impl import random_rho_pure as _random_rho_pure
from .types import complex_dtypes

# ~~ Exports ~~ ────────────────────────────────────────────────────────────────
__all__: List[str] = [
    "random_rho_bh",
    "random_rho_hs",
    "random_obs_csu",
    "random_obs_gue",
    "random_obs_cgi",
    "random_rho_pure",
]


# ~~ Functions ~~ ──────────────────────────────────────────────────────────────
# noinspection DuplicatedCode
def random_rho_hs(
    size: int,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Optional[Tuple[int, ...]] = None,
) -> Tensor:
    """
    Generate a random quantum state (or a batch thereof) uniformly w.r.t. the Hilbert-Schmidt measure.

    Parameters
    ----------
    size : int
        The size of the square matrix.
    dtype : torch.dtype
        The data type. Default is torch.double.
    device : torch.device, optional
        The device. Default is None.
    batch_shape : tuple of ints, optional
        The batch shape for generating multiple matrices. Default is None.

    Returns
    -------
    Tensor
        A random quantum state (or a batch thereof) uniformly w.r.t. the Hilbert-Schmidt measure.
    """
    check_size(size)
    check_dtype(dtype, complex_dtypes)
    bs = () if batch_shape is None else batch_shape
    return _random_rho_hs(size=size, dtype=dtype, device=device, batch_shape=bs)


def random_rho_bh(
    size: int,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Optional[Tuple[int, ...]] = None,
) -> Tensor:
    """
    Generate a random quantum state (or a batch thereof) uniformly w.r.t. the Bures-Helstrom measure.

    Parameters
    ----------
    size : int
        The size of the square matrix.
    dtype : torch.dtype
        The data type. Default is torch.double.
    device : torch.device, optional
        The device. Default is None.
    batch_shape : tuple of ints, optional
        The batch shape for generating multiple matrices. Default is None.

    Returns
    -------
    Tensor
        A random quantum state (or a batch thereof) uniformly w.r.t. the Bures-Helstrom measure.
    """
    check_size(size)
    check_dtype(dtype, complex_dtypes)
    bs = () if batch_shape is None else batch_shape
    return _random_rho_bh(size=size, dtype=dtype, device=device, batch_shape=bs)


def random_obs_csu(
    size: int,
    evdist: Callable[..., Tensor],
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Optional[Tuple[int, ...]] = None,
) -> Tensor:
    """
    Generate a random observable (or a batch thereof) uniformly in a compact set from the distribution of eigenvalues.

    Parameters
    ----------
    size : int
        The size of the square matrix.
    evdist : Callable[..., Tensor]
        A function to sample from distribution of eigenvalues. Must implement the Torch API.
    dtype : torch.dtype
        The data type. Default is torch.double.
    device : torch.device, optional
        The device. Default is None.
    batch_shape : tuple of ints, optional
        The batch shape for generating multiple matrices. Default is None.

    Returns
    -------
    Tensor
        A random observable (or a batch thereof) uniformly w.r.t. the CSU measure.
    """
    check_size(size)
    check_dtype(dtype, complex_dtypes)
    bs = () if batch_shape is None else batch_shape
    return _random_obs_csu(
        size=size, evdist=evdist, dtype=dtype, device=device, batch_shape=bs
    )


def random_obs_gue(
    size: int,
    sigma: float,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Optional[Tuple[int, ...]] = None,
) -> Tensor:
    """
    Generate a random observable (or a batch thereof) uniformly in the Gaussian Unitary Ensemble.

    Parameters
    ----------
    size : int
        The size of the square matrix.
    sigma : float
        The scale parameter.
    dtype : torch.dtype
        The data type. Default is torch.double.
    device : torch.device, optional
        The device. Default is None.
    batch_shape : tuple of ints, optional
        The batch shape for generating multiple matrices. Default is None.

    Returns
    -------
    Tensor
        A random observable (or a batch thereof) uniformly w.r.t. the GUE measure.
    """
    check_size(size)
    check_dtype(dtype, complex_dtypes)
    bs = () if batch_shape is None else batch_shape
    return _random_obs_gue(
        size=size, sigma=sigma, dtype=dtype, device=device, batch_shape=bs
    )


def random_obs_cgi(
    size: int,
    coeff_low: float = 0.0,
    coeff_upp: float = 1.0,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Optional[Tuple[int, ...]] = None,
) -> Tensor:
    """
    Generate a random observable (or a batch thereof) uniformly in the Gell-Mann parameterization of Hermitians.

    Parameters
    ----------
    size : int
        The size of the square matrix.
    coeff_low : float, optional
        The lower bound for the coefficients. Default is 0.0.
    coeff_upp : float, optional
        The upper bound for the coefficients. Default is 1.0.
    dtype : torch.dtype
        The data type. Default is torch.double.
    device : torch.device, optional
        The device. Default is None.
    batch_shape : tuple of ints, optional
        The batch shape for generating multiple matrices. Default is None.

    Returns
    -------
    Tensor
        A random observable (or a batch thereof) uniformly w.r.t. the CGI measure.
    """
    check_size(size, disallow_equality=True)
    check_bounds(coeff_low, coeff_upp)
    check_dtype(dtype, complex_dtypes)
    bs = () if batch_shape is None else batch_shape
    return _random_obs_cgi(
        size=size,
        coeff_low=coeff_low,
        coeff_upp=coeff_upp,
        dtype=dtype,
        device=device,
        batch_shape=bs,
    )


def random_rho_pure(
    size: int,
    dtype: th.dtype = th.cdouble,
    device: Optional[th.device] = None,
    batch_shape: Optional[Tuple[int, ...]] = None,
    *,
    bo_einsum: bool = False,
):
    # TODO: Write docs!
    check_size(size)
    check_dtype(dtype, complex_dtypes)
    bs = () if batch_shape is None else batch_shape
    return _random_rho_pure(
        size=size,
        dtype=dtype,
        device=device,
        batch_shape=bs,
        bo_einsum=bo_einsum,
    )
